@@ -17,10 +17,12 @@ import os
 import re
 import sys
 import json
+import time
 import stat
 import errno
 import signal
 import logging
+import tempfile
 import subprocess
 from datetime import datetime, timedelta
 
@@ -30,6 +32,17 @@ from salttesting.helpers import RedirectStdStreams
 from salttesting.runtests import RUNTIME_VARS
 from salttesting.mixins import AdaptedConfigurationTestCaseMixIn, SaltClientTestCaseMixIn
 
+# Try to import salt: needed for __salt_system_encoding__ reference
+try:
+    current_module_names = sys.modules.keys()
+    import salt
+
+    for name in sys.modules.keys():
+        if name not in current_module_names:
+            del sys.modules[name]
+except ImportError:
+    pass
+
 STATE_FUNCTION_RUNNING_RE = re.compile(
     r'''The function (?:"|')(?P<state_func>.*)(?:"|') is running as PID '''
     r'(?P<pid>[\d]+) and was started at (?P<date>.*) with jid (?P<jid>[\d]+)'
@@ -97,7 +110,7 @@ class ShellTestCase(TestCase, AdaptedConfigurationTestCaseMixIn):
 
         return script_path
 
-    def run_salt(self, arg_str, with_retcode=False, catch_stderr=False):
+    def run_salt(self, arg_str, with_retcode=False, catch_stderr=False, timeout=15):
         r'''
         Run the ``salt`` CLI tool with the provided arguments
 
@@ -115,7 +128,7 @@ class ShellTestCase(TestCase, AdaptedConfigurationTestCaseMixIn):
         arg_str = '-c {0} {1}'.format(self.get_config_dir(), arg_str)
         return self.run_script('salt', arg_str, with_retcode=with_retcode, catch_stderr=catch_stderr)
 
-    def run_ssh(self, arg_str, with_retcode=False, catch_stderr=False):
+    def run_ssh(self, arg_str, with_retcode=False, timeout=25, catch_stderr=False):
         '''
         Execute salt-ssh
         '''
@@ -137,26 +150,46 @@ class ShellTestCase(TestCase, AdaptedConfigurationTestCaseMixIn):
                                                  async_flag=' --async' if async else '')
         return self.run_script('salt-run', arg_str, with_retcode=with_retcode, catch_stderr=catch_stderr)
 
-    def run_run_plus(self, fun, options='', *arg, **kwargs):
+    def run_run_plus(self, fun, *arg, **kwargs):
         '''
-        Execute Salt run and the salt run function and return the data from
-        each in a dict
+        Execute the runner function and return the return data and output in a dict
         '''
-        ret = {}
-        ret['out'] = self.run_run(
-            '{0} {1} {2}'.format(options, fun, ' '.join(arg)), catch_stderr=kwargs.get('catch_stderr', None)
-        )
+        ret = {'fun': fun}
+
         # Late import
         import salt.config
+        import salt.output
         import salt.runner
+        from salt.ext.six.moves import cStringIO
 
         opts = salt.config.master_config(
             self.get_config_file_path('master')
         )
-        opts.update({'doc': False, 'fun': fun, 'arg': arg})
+
+        opts_arg = list(arg)
+        if kwargs:
+            opts_arg.append({'__kwarg__': True})
+            opts_arg[-1].update(kwargs)
+
+        opts.update({'doc': False, 'fun': fun, 'arg': opts_arg})
         with RedirectStdStreams():
             runner = salt.runner.Runner(opts)
-            ret['fun'] = runner.run()
+            ret['return'] = runner.run()
+            try:
+                ret['jid'] = runner.jid
+            except AttributeError:
+                ret['jid'] = None
+
+        # Compile output
+        # TODO: Support outputters other than nested
+        opts['color'] = False
+        opts['output_file'] = cStringIO()
+        try:
+            salt.output.display_output(ret['return'], opts=opts)
+            ret['out'] = opts['output_file'].getvalue()
+        finally:
+            opts['output_file'].close()
+
         return ret
 
     def run_key(self, arg_str, catch_stderr=False, with_retcode=False):
@@ -194,7 +227,8 @@ class ShellTestCase(TestCase, AdaptedConfigurationTestCaseMixIn):
                    arg_str,
                    catch_stderr=False,
                    with_retcode=False,
-                   timeout=None,
+                   # FIXME A timeout of zero or disabling timeouts may not return results!
+                   timeout=15,
                    raw=False):
         '''
         Execute a script with the given argument string
@@ -204,18 +238,28 @@ class ShellTestCase(TestCase, AdaptedConfigurationTestCaseMixIn):
             return False
 
         python_path = os.environ.get('PYTHONPATH', None)
-        cmd = 'PYTHONPATH='
+
+        if sys.platform.startswith('win'):
+            cmd = 'set PYTHONPATH='
+        else:
+            cmd = 'PYTHONPATH='
+
         if python_path is not None:
             cmd += '{0}:'.format(python_path)
 
-        cmd += '{0} '.format(':'.join(sys.path[1:]))
+        if sys.version_info[0] < 3:
+            cmd += '{0} '.format(':'.join(sys.path[1:]))
+        else:
+            cmd += '{0} '.format(':'.join(sys.path[0:]))
         cmd += 'python{0}.{1} '.format(*sys.version_info)
         cmd += '{0} '.format(script_path)
         cmd += '{0} '.format(arg_str)
 
+        tmp_file = tempfile.SpooledTemporaryFile()
+
         popen_kwargs = {
             'shell': True,
-            'stdout': subprocess.PIPE,
+            'stdout': tmp_file,
             'universal_newlines': True
         }
 
@@ -241,7 +285,7 @@ class ShellTestCase(TestCase, AdaptedConfigurationTestCaseMixIn):
             term_sent = False
             while True:
                 process.poll()
-
+                time.sleep(0.1)
                 if datetime.now() > stop_at:
                     if term_sent is False:
                         # Kill the process group since sending the term signal
@@ -275,6 +319,16 @@ class ShellTestCase(TestCase, AdaptedConfigurationTestCaseMixIn):
 
                 if process.returncode is not None:
                     break
+        tmp_file.seek(0)
+
+        if sys.version_info >= (3,):
+            try:
+                out = tmp_file.read().decode(__salt_system_encoding__)
+            except (NameError, UnicodeDecodeError):
+                # Let's cross our fingers and hope for the best
+                out = tmp_file.read().decode('utf-8')
+        else:
+            out = tmp_file.read()
 
         if catch_stderr:
             if sys.version_info < (2, 7):
@@ -291,10 +345,9 @@ class ShellTestCase(TestCase, AdaptedConfigurationTestCaseMixIn):
                 #
                 # Use this work around were it's needed only, python 2.6
                 process.wait()
-                out = process.stdout.read()
                 err = process.stderr.read()
             else:
-                out, err = process.communicate()
+                _, err = process.communicate()
             # Force closing stderr/stdout to release file descriptors
             if process.stdout is not None:
                 process.stdout.close()
@@ -327,22 +380,32 @@ class ShellTestCase(TestCase, AdaptedConfigurationTestCaseMixIn):
                     pass
             # pylint: enable=maybe-no-member
 
-        data = process.communicate()
-        process.stdout.close()
+        # TODO Remove this?
+        process.communicate()
+        if process.stdout is not None:
+            process.stdout.close()
 
         try:
             if with_retcode:
                 if not raw:
-                    return data[0].splitlines(), process.returncode
+                    return out.splitlines(), process.returncode
                 else:
-                    return data[0], process.returncode
+                    return out, process.returncode
             else:
                 if not raw:
-                    return data[0].splitlines()
+                    return out.splitlines()
                 else:
-                    return data[0]
+                    return out
         finally:
             try:
+                if os.path.exists(tmp_file.name):
+                    if isinstance(tmp_file.name, str):
+                        # tmp_file.name is an int when using SpooledTemporaryFiles
+                        # int types cannot be used with os.remove() in Python 3
+                        os.remove(tmp_file.name)
+                    else:
+                        # Clean up file handles
+                        tmp_file.close()
                 process.terminate()
             except OSError as err:
                 # process already terminated
