@@ -15,7 +15,10 @@
 from __future__ import absolute_import
 import os
 import sys
+import time
+import errno
 import types
+import signal
 import socket
 import inspect
 import logging
@@ -27,8 +30,9 @@ from salttesting.unit import skip, _id
 
 # Import 3rd-party libs
 import six
-if sys.version_info < (3,):
-    import __builtin__  # pylint: disable=incompatible-py3-code
+import psutil
+if six.PY2:
+    import __builtin__  # pylint: disable=incompatible-py3-code,import-error
 else:
     import builtins as __builtin__  # pylint: disable=import-error
 
@@ -103,6 +107,38 @@ def expensiveTest(caller):
     return wrap
 
 
+def flaky(caller):
+    '''
+    Mark a test as flaky. The test will attempt to run five times,
+    looking for a successful run. After an immediate second try,
+    it will use an exponential backoff starting with one second.
+
+    .. code-block:: python
+
+        class MyTestCase(TestCase):
+
+        @flaky
+        def test_sometimes_works(self):
+            pass
+    '''
+    # TODO Handle class case. Right now, just POC on a function
+
+    @wraps(caller)
+    def wrap(cls):
+        for attempt in range(0, 4):
+            try:
+                caller(cls)
+                break
+            except Exception as exc:
+                if attempt == 4:
+                    raise exc
+                backoff_time = attempt ** 2
+                print('Found exception. Waiting {0} seconds to retry.'.format(backoff_time))
+                time.sleep(backoff_time)
+        return None
+    return wrap
+
+
 def requires_sshd_server(caller):
     '''
     Mark a test as requiring the tests SSH daemon running.
@@ -131,7 +167,7 @@ def requires_sshd_server(caller):
     @wraps(caller)
     def wrap(cls):
         if os.environ.get('SSH_DAEMON_RUNNING', 'False').lower() == 'false':
-            self.skipTest('SSH tests are disabled')
+            cls.skipTest('SSH tests are disabled')
         return caller(cls)
     return wrap
 
@@ -1056,3 +1092,164 @@ def skip_if_not_root(func):
         func.__unittest_skip__ = True
         func.__unittest_skip_why__ = 'You must be logged in as root to run this test'
     return func
+
+
+if sys.platform.startswith('win'):
+    SIGTERM = signal.CTRL_BREAK_EVENT  # pylint: disable=no-member
+else:
+    SIGTERM = signal.SIGTERM
+
+
+def collect_child_processes(pid):
+    '''
+    Try to collect any started child processes of the provided pid
+    '''
+    # Let's get the child processes of the started subprocess
+    try:
+        parent = psutil.Process(pid)
+        if hasattr(parent, 'children'):
+            children = parent.children(recursive=True)
+        else:
+            children = []
+    except psutil.NoSuchProcess:
+        children = []
+    return children[::-1]  # return a reversed list of the children
+
+
+def _terminate_process_list(process_list, kill=False, slow_stop=False):
+    for process in process_list[:][::-1]:  # Iterate over a reversed copy of the list
+        if not psutil.pid_exists(process.pid):
+            process_list.remove(process)
+            continue
+        try:
+            if not kill and process.status() == psutil.STATUS_ZOMBIE:
+                # Zombie processes will exit once child processes also exit
+                continue
+            try:
+                cmdline = process.cmdline()
+            except psutil.AccessDenied:
+                # OSX is more restrictive about the above information
+                cmdline = None
+            if not cmdline:
+                try:
+                    cmdline = process.as_dict()
+                except Exception:
+                    cmdline = 'UNKNOWN PROCESS'
+            if kill:
+                log.info('Killing process(%s): %s', process.pid, cmdline)
+                process.kill()
+            else:
+                log.info('Terminating process(%s): %s', process.pid, cmdline)
+                try:
+                    if slow_stop:
+                        # Allow coverage data to be written down to disk
+                        process.send_signal(SIGTERM)
+                        try:
+                            process.wait(2)
+                        except psutil.TimeoutExpired:
+                            if psutil.pid_exists(process.pid):
+                                continue
+                    else:
+                        process.terminate()
+                except OSError as exc:
+                    if exc.errno not in (errno.ESRCH, errno.EACCES):
+                        raise
+            if not psutil.pid_exists(process.pid):
+                process_list.remove(process)
+        except psutil.NoSuchProcess:
+            process_list.remove(process)
+
+
+def terminate_process_list(process_list, kill=False, slow_stop=False):
+
+    def on_process_terminated(proc):
+        log.info('Process %s terminated with exit code: %s', getattr(proc, '_cmdline', proc), proc.returncode)
+
+    # Try to terminate processes with the provided kill and slow_stop parameters
+    log.info('Terminating process list. 1st step. kill: %s, slow stop: %s', kill, slow_stop)
+
+    # Cache the cmdline since that will be inaccessible once the process is terminated
+    for proc in process_list:
+        try:
+            cmdline = proc.cmdline()
+        except (psutil.NoSuchProcess, psutil.AccessDenied):
+            # OSX is more restrictive about the above information
+            cmdline = None
+        if not cmdline:
+            try:
+                cmdline = proc
+            except (psutil.NoSuchProcess, psutil.AccessDenied):
+                cmdline = '<could not be retrived; dead process: {0}>'.format(proc)
+        proc._cmdline = cmdline
+    _terminate_process_list(process_list, kill=kill, slow_stop=slow_stop)
+    psutil.wait_procs(process_list, timeout=15, callback=on_process_terminated)
+
+    if process_list:
+        # If there's still processes to be terminated, retry and kill them if slow_stop is False
+        log.info('Terminating process list. 2nd step. kill: %s, slow stop: %s', slow_stop is False, slow_stop)
+        _terminate_process_list(process_list, kill=slow_stop is False, slow_stop=slow_stop)
+        psutil.wait_procs(process_list, timeout=10, callback=on_process_terminated)
+
+    if process_list:
+        # If there's still processes to be terminated, just kill them, no slow stopping now
+        log.info('Terminating process list. 3rd step. kill: True, slow stop: False')
+        _terminate_process_list(process_list, kill=True, slow_stop=False)
+        psutil.wait_procs(process_list, timeout=5, callback=on_process_terminated)
+
+    if process_list:
+        # In there's still processes to be terminated, log a warning about it
+        log.warning('Some processes failed to properly terminate: %s', process_list)
+
+
+def terminate_process(pid=None, process=None, children=None, kill_children=False, slow_stop=False):
+    '''
+    Try to terminate/kill the started processe
+    '''
+    children = children or []
+    process_list = []
+
+    def on_process_terminated(proc):
+        log.info('Process %s terminated with exit code: %s', getattr(proc, '_cmdline', proc), proc.returncode)
+
+    if pid and not process:
+        try:
+            process = psutil.Process(pid)
+            process_list.append(process)
+        except psutil.NoSuchProcess:
+            # Process is already gone
+            process = None
+
+    if kill_children:
+        if process:
+            if not children:
+                children = collect_child_processes(process.pid)
+            else:
+                # Let's collect children again since there might be new ones
+                children.extend(collect_child_processes(pid))
+        if children:
+            process_list.extend(children)
+
+    if process_list:
+        if process:
+            log.info('Stopping process %s and respective children: %s', process, children)
+        else:
+            log.info('Terminating process list: %s', process_list)
+        terminate_process_list(process_list, kill=slow_stop is False, slow_stop=slow_stop)
+        if psutil.pid_exists(pid):
+            log.warning('Process left behind which we were unable to kill: %s', process)
+
+
+def terminate_process_pid(pid, only_children=False):
+    children = []
+    process = None
+
+    # Let's begin the shutdown routines
+    try:
+        process = psutil.Process(pid)
+        children = collect_child_processes(pid)
+    except psutil.NoSuchProcess:
+        log.info('No process with the PID %s was found running', pid)
+
+    if only_children:
+        return terminate_process(children=children, kill_children=True, slow_stop=True)
+    return terminate_process(pid=pid, process=process, children=children, kill_children=True, slow_stop=True)
